from enum import Enum

class Anchor(Enum):
    LEFT_TOP = ("left" ,"top")
    LEFT_CENTER = ("left" ,"center")
    LEFT_BOTTOM = ("left" ,"bottom")
    CENTER_TOP = ("center" ,"top")
    CENTER = ("center" ,"center")
    CENTER_BOTTOM = ("center" ,"bottom")
    RIGHT_TOP = ("right" ,"top")
    RIGHT_CENTER = ("right" ,"center")
    RIGHT_BOTTOM = ("right" ,"bottom")


from .base import wx

def OpenDirectory(title="Open Directory", directory=""):
    with wx.DirDialog(None, title, directory, wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST) as dir_dialog:
        if dir_dialog.ShowModal() == wx.ID_CANCEL:
            return None
        return dir_dialog.GetPath()

def OpenFile(title="Open File", directory="", wildcard=wx.FileSelectorDefaultWildcardStr):
    with wx.FileDialog(None, title, directory, wildcard=wildcard,
                       style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as file_dialog:
        if file_dialog.ShowModal() == wx.ID_CANCEL:
            return None
        return file_dialog.GetPath()

def OpenFiles(title="Open Files", directory="", wildcard=wx.FileSelectorDefaultWildcardStr):
    with wx.FileDialog(None, title, directory, wildcard=wildcard,
                       style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.FD_MULTIPLE) as file_dialog:
        if file_dialog.ShowModal() == wx.ID_CANCEL:
            return None
        return file_dialog.GetPaths()

def SaveFile(default, title="Save File", directory="", wildcard=wx.FileSelectorDefaultWildcardStr):
    if not directory and isinstance(default, str):
        directory = default
    with wx.FileDialog(None, title, directory, wildcard=wildcard,
                       style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as file_dialog:
        if file_dialog.ShowModal() == wx.ID_CANCEL:
            return None
        return file_dialog.GetPath()

def Information(message="Information", title="Information Dialog"):
    wx.MessageBox(message, title, wx.OK | wx.ICON_INFORMATION)

def Warning(message="Warning", title="Warning Dialog"):
    wx.MessageBox(message, title, wx.OK | wx.ICON_WARNING)

def Critical(message="Critical", title="Critical Dialog"):
    wx.MessageBox(message, title, wx.OK | wx.ICON_ERROR)

def Confirm(message="Confirm", title="Confirm Dialog"):
    dlg = wx.MessageDialog(None, message, title, wx.YES_NO | wx.ICON_QUESTION)
    result = dlg.ShowModal()
    dlg.Destroy()
    return result == wx.ID_YES

def Prompt(prompt="Input", title="Input Dialog", default=""):
    dlg = wx.TextEntryDialog(None, prompt, title, default)
    if dlg.ShowModal() == wx.ID_OK:
        result = dlg.GetValue()
    else:
        result = None
    dlg.Destroy()
    return result


# dprint = print
dprint = lambda *x: x

def recur_delete(node, child, direct):
    child.destroyed = True
    if not child.pui_virtual:
        for sc in child.children:
            recur_delete(child, sc, False)
    dprint("recur_delete", node.key, child.key, direct)
    child.destroy(direct)

def sortGridDOM(dom):
    dom = [c for c in dom if c.grid_row is not None and c.grid_column is not None]
    return sorted(dom, key=lambda c:(c.grid_row, c.grid_column, c.grid_rowspan, c.grid_columnspan))

def remove_node(dom_parent, offset, child):
    if child.pui_virtual:
        ret = [child]
        for c in child.children:
            ret.extend(remove_node(dom_parent, offset, c))
        return ret
    else:
        dom_parent.removeChild(offset , child)
        return [child]

def add_nodes(dom_parent, offset, children):
    for i, c in enumerate(children):
        dom_parent.addChild(offset+i, c)

def sync(node, dom_parent, offset, oldDOM, newDOM):
    dprint(f"Syncing {node.key} parent={dom_parent.key} offset={offset} old={len(oldDOM)} new={len(newDOM)}")

    if node.pui_grid_layout:
        oldDOM = sortGridDOM(oldDOM)
        newDOM = sortGridDOM(newDOM)

    dprint("  ===OLD===")
    for c in oldDOM:
        dprint(f"   {c.key} virtual={c.pui_virtual} ui={c.ui}")

    dprint("  ===NEW===")
    for c in newDOM:
        dprint(f"   {c.key} virtual={c.pui_virtual}")

    oldMap = [x.key for x in oldDOM]
    newMap = [x.key for x in newDOM]

    node.preSync()

    tbd = []
    for i, new in enumerate(newDOM):
        dprint(f"sync child {i}, {new.key} dom_parent={dom_parent.key} virtual={new.pui_virtual}")
        new.pui_vparent = dom_parent if dom_parent is not node else None

        while True:
            # Step 1. just matched
            if i < len(oldDOM) and oldMap[i] == new.key: # matched
                dprint(f"MATCHED {i} {new.key}")
                old = oldDOM[i]

                try:
                    new.update(old)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(new.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")

                if new.pui_virtual:
                    offset = sync(new, dom_parent, offset+i, old.children, new.children) - i - 1
                else:
                    if new.pui_outoforder:
                        offset -= 1

                    if not new.pui_terminal:
                        sync(new, new, 0, old.children, new.children)

                break # finish

            # Step 2. trim removed nodes after common prefix
            trimmed = False
            while i < len(oldDOM) and not oldDOM[i].key in newMap[i:]: # trim old nodes
                dprint(f"TRIM {i} {oldDOM[i].key}")
                old = oldDOM.pop(i)
                oldMap.pop(i)
                tbd.extend(remove_node(dom_parent, offset+i, old))
                trimmed = True

            if trimmed:
                continue # restart

            # Step 3. setup target node

            try:
                idx = oldMap[i+1:].index(new.key)+i+1
            except ValueError:
                idx = None

            # Step 3-1. new node
            if idx is None:
                try:
                    new.update(None)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(new.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")

                if new.pui_virtual:
                    offset = sync(new, dom_parent, offset+i, [], new.children) - i - 1
                else:
                    dprint("addChild", dom_parent.key, offset+i, new.key)
                    dom_parent.addChild(offset+i, new)

                    if new.pui_outoforder:
                        offset -= 1

                    if not new.pui_terminal:
                        sync(new, new, 0, [], new.children)

                oldDOM.insert(i, None) # placeholder
                oldMap.insert(i, new.key)

            # Step 3-2. existed node
            else:
                # if the target node is in just next position, requeue the front to prevent repositioning every nodes coming after
                # eg. when an element is removed from a long list, do single 3-2-1 instead of many 3-2-2

                # Step 3-2-1. yield the next position for the target node
                if idx==i+1:
                    oldMap.pop(i)
                    old = oldDOM.pop(i)
                    nodes = remove_node(dom_parent, offset+i, old)
                    add_nodes(dom_parent, offset+len(oldDOM), nodes)
                    oldMap.append(old.key)
                    oldDOM.append(old)

                    # sync will be peformed in next step 1

                    continue # restart

                 # Step 3-2-2. move target node
                else:
                    oldMap.pop(idx)
                    old = oldDOM.pop(idx)

                    if old.pui_virtual:
                        nodes = remove_node(dom_parent, offset+idx, old)
                        add_nodes(dom_parent, offset+i, nodes)
                        offset = sync(new, dom_parent, offset+i, old.children, new.children) - i - 1
                    else:
                        dom_parent.removeChild(offset+idx, old)

                        try:
                            new.update(old)
                        except:
                            import traceback
                            print("## <ERROR OF update() >")
                            print(new.key)
                            traceback.print_exc()
                            print("## </ERROR OF update()>")

                        dom_parent.addChild(offset+i, new)
                        if new.pui_outoforder:
                            offset -= 1

                        if not new.pui_terminal:
                            sync(new, new, 0, old.children, new.children)

                    oldDOM.insert(i, None) # placeholder
                    oldMap.insert(i, new.key)

            break # finish

    # Step 4. trim removed trail
    nl = len(newDOM)
    while len(oldDOM) > nl:
        old = oldDOM.pop(nl)
        oldMap.pop(nl)
        tbd.extend(remove_node(dom_parent, offset+nl, old))

    for c in newDOM:
        c.postUpdate()

    node.postSync()

    # release deleted nodes
    for old in tbd:
        recur_delete(node, old, True)

    dprint(f"sync end -> {offset+nl}")
    return offset+nl


from .. import *
from .base import *
import itertools
class Canvas(TkBaseWidget):
    pui_terminal = True
    def __init__(self, painter, *args):
        super().__init__()
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = tk.Canvas(self.tkparent.inner)
        self.ui.delete("all")

        self.painter(self, *self.args)
        super().update(prev)

    def drawText(self, x, y, text, w=None, h=None, rotate=0, anchor=Anchor.LEFT_TOP):
        if rotate !=0:
            print("drawText: rotate not implemented")
        self.ui.create_text(x, y, text=text)

    def drawLine(self, x1, y1, x2, y2, color=None, width=None):
        params = {}
        if not color is None:
            params["fill"] = f"#{color:06X}"
        if not width is None:
            params["width"] = width
        self.ui.create_line(x1, y1, x2, y2, **params)

    def drawPolyline(self, coords, color=None, width=None):
        params = {}
        if not color is None:
            params["fill"] = f"#{color:06X}"
        if not width is None:
            params["width"] = width
        self.ui.create_line(*itertools.chain(*coords), **params)

    def drawPolygon(self, coords, fill=None, stroke=None, width=1):
        print("drawPolygon not implemented")

    def drawRect(self, x1, y1, x2, y2, fill=None, stroke=None, width=1):
        print("drawRect not implemented")

    def drawEllipse(self, x, y, rx, ry, fill=None, stroke=None, width=1):
        print("drawEllipse not implemented")



from .. import *
from .base import *

class TextField(FBase):
    def __init__(self, model, edit_model=None, label="", **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.label = label
        self.edit_model = edit_model
        self.editing = False

    def update(self, prev):
        model_value = str(self.model.value)
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.curr_value = prev.curr_value
            if self.curr_value.set(model_value):
                self.ui.value = model_value
                try:
                    self.ui.update()
                except:
                    pass
        else:
            self.curr_value = Prop(model_value)
            self.ui = ft.TextField(label=self.label, value=model_value, on_change=self.on_textbox_changed, on_blur=self.on_blur, on_submit=self.on_submit, expand=self.layout_weight)

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        super().update(prev)

    def on_textbox_changed(self, e):
        node = self.get_node()
        node.editing = True
        value = e.control.value
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._input(e)

    def on_change(self, e):
        node = self.get_node()
        node.editing = False
        value = e.control.value
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._change(e)

    def on_blur(self, e):
        self.on_change(e)

    def on_submit(self, e):
        self.on_change(e)



from .. import *
from .base import *

class TextField(TBase):
    content_width = None
    def __init__(self, model, edit_model=None):
        super().__init__()
        self.model = model
        self.edit_model = edit_model
        self.editing = False

    def update(self, prev):
        model_value = str(self.model.value)
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.curr_value = prev.curr_value

            if self.curr_value.set(model_value) and not self.editing:
                self.ui.value = model_value
        else:
            self.ui = widgets.Input(model_value)
            self.curr_value = Prop(model_value)

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        self.ui.puinode = self
        super().update(prev)


    def _changed(self, value):
        node = self.get_node()
        node.editing = True
        if node.edit_model:
           node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._input(e)

    def _submitted(self, value):
        node = self.get_node()
        node.editing = False
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._change(e)


from .. import *
import flet as ft

def _apply_params(ui, params):
    for k,v in params.items():
        setattr(ui, k, v)

class FView(PUIView):
    pui_virtual = True

class FBase(PUINode):
    def __init__(self, *args):
        super().__init__(*args)
        self.child_weight = None
        if self.fparent:
            self.layout_weight = self.fparent.child_weight

        self.flet_params = {}

    def flet(self, **kwargs):
        for k,v in kwargs.items():
            self.flet_params[k] = v
        return self

    def update(self, prev):
        _apply_params(self.ui, self.flet_params)
        super().update(prev)

    @property
    def fparent(self):
        parent = self.parent
        while True:
            if isinstance(parent, FBase):
                return parent
            if parent.parent == parent:
                return None
            parent = parent.parent



from .. import *
from .base import *

class HBox(WxBaseLayout):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = wx.BoxSizer(wx.HORIZONTAL)
        super().update(prev)

class VBox(WxBaseLayout):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = wx.BoxSizer(wx.VERTICAL)
        super().update(prev)

class Spacer(PUINode):
    pui_terminal = True
    def __init__(self):
        super().__init__()
        self.layout_weight = 1

class Grid(WxBaseLayout):
    pui_grid_layout = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = wx.GridBagSizer()
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, WxBaseLayout) or isinstance(child, WxBaseWidget):
            self.ui.Add(child.outer, pos=(child.grid_row, child.grid_column), span=(child.grid_rowspan or 1, child.grid_columnspan or 1), flag=wx.ALL)

    def removeChild(self, idx, child):
        if isinstance(child, WxBaseLayout) or isinstance(child, WxBaseWidget):
            pass



from .. import *
from .base import *

class Application(WxPUIView):
    def __init__(self):
        super().__init__()
        self.ui = None

    def redraw(self):
        if self.ui:
            super().redraw()
        else:
            self.update(None)

    def update(self, prev=None):
        if not self.ui:
            self.ui = wx.App()

        super().update(prev)

    def addChild(self, idx, child):
        child.outer.Show()

    def removeChild(self, idx, child):
        child.outer.Close()

    def start(self):
        self.ui.MainLoop()

def PUIApp(func):
    def func_wrapper(*args, **kwargs):
        class PUIAppWrapper(Application):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIAppWrapper(func.__name__)
        return ret

    return func_wrapper



from .. import *
from .base import *
from .menu import *
from .modal import *
from .toolbar import *
from PySide6 import QtWidgets

class QMainWindow(QtWidgets.QMainWindow):
    def keyPressEvent(self, event):
        e = PUIEvent()
        e.text = event.text()
        self.node._keypress(e)

    def mousePressEvent(self, event):
        focused_widget = QtWidgets.QApplication.focusWidget()
        if isinstance(focused_widget, QtWidgets.QLineEdit):
            focused_widget.clearFocus()
        super().mousePressEvent(event)

class Window(QtBaseWidget):
    pui_terminal = False

    def __init__(self, title=None, icon=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        if icon:
            self.icon = QtGui.QIcon(icon)
        else:
            self.icon = None
        self.size = size
        self.maximize = maximize
        self.fullscreen = fullscreen

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.node = self
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
        else:
            self.ui = QMainWindow()
            self.ui.node = self
            self.ui.show()
            self.curr_size = Prop()
            self.curr_maximize = Prop()
            self.curr_fullscreen = Prop()

        if self.curr_size.set(self.size):
            self.ui.resize(*self.size)
        if self.curr_maximize.set(self.maximize):
            self.ui.showMaximized()
        if self.curr_fullscreen.set(self.fullscreen):
            self.ui.showFullScreen()
        if not self.title is None:
            self.ui.setWindowTitle(self.title)
        if not self.icon is None:
            self.ui.setWindowIcon(self.icon)
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, MenuBar):
            self.ui.setMenuBar(child.outer)
        elif isinstance(child, ToolBar):
            self.ui.addToolBar(child.outer)
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.ui.setCentralWidget(child.outer)
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, MenuBar):
            child.outer.close()
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
        else:
            self.removeChild(idx, child.children[0])



from .. import *

import wx

def int_to_wx_colour(color_int):
    red = (color_int >> 16) & 0xFF
    green = (color_int >> 8) & 0xFF
    blue = color_int & 0xFF
    return wx.Colour(red, green, blue)

def getWindow(n):
    from .window import Window
    from .scroll import Scroll
    p = n
    while p:
        if isinstance(p, Scroll):
            return p.ui
        if isinstance(p, Window):
            return p.ui
        p = p.parent
    raise RuntimeError("!!! getWindow returns None for", n.key if n else None)

class WxPUIView(PUIView):
    pui_virtual = True
    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        if self.ui: # PUIView doesn't have ui
            self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def redraw(self):
        wx.CallAfter(self.sync)

    def update(self, prev=None):
        if self.retired_by:
            return
        self.dirty = False
        super().update(prev)
        self.updating = False
        if self.dirty:
            self.update(prev)

class WxBaseWidget(PUINode):
    pui_terminal = True

    def __init__(self):
        super().__init__()

    def update(self, prev):
        super().update(prev)

        if self.style_fontsize:
            font = self.ui.GetFont()
            font.SetPointSize(self.style_fontsize)
            self.ui.SetFont(font)

        # Set font color if specified
        if self.style_color:
            self.ui.SetForegroundColour(int_to_wx_colour(self.style_color))

    def destroy(self, direct):
        self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

class WxBaseLayout(PUINode):
    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        # self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def addChild(self, idx, child):
        from .layout import Spacer
        weight = child.layout_weight
        if weight is None:
            weight = 0
        flag = wx.EXPAND|wx.ALL
        if isinstance(child, WxBaseLayout):
            self.ui.Insert(idx, child.outer, weight, flag)
        elif isinstance(child, WxBaseWidget):
            self.ui.Insert(idx, child.outer, weight, flag)
        elif isinstance(child, Spacer):
            self.ui.InsertStretchSpacer(idx, weight)

    def removeChild(self, idx, child):
        from .layout import Spacer
        if isinstance(child, WxBaseLayout):
            self.ui.Detach(idx)
        elif isinstance(child, WxBaseWidget):
            self.ui.Detach(idx)
        elif isinstance(child, Spacer):
            self.ui.Detach(idx)



from .. import *
from .base import *

class Tabs(TkBaseWidget):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    pui_terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition

    def destroy(self, direct):
        if direct:
            for w in self.widgets:
                if w:
                    w.deleteLater()
        super().destroy(direct)

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.childrenui = prev.childrenui
        else:
            self.ui = ttk.Notebook(self.tkparent.inner)
            self.childrenui = []
        super().update(prev)

    def addChild(self, idx, child):
        if idx < len(self.childrenui):
            self.ui.insert(self.childrenui[idx], child.outer, text=child.parent.label)
            self.childrenui.insert(idx, child.outer)
        else:
            self.ui.add(child.outer, text=child.parent.label)
            self.childrenui.append(child.outer)

    def removeChild(self, idx, child):
        self.ui.forget(child.outer)
        self.childrenui.pop(idx)

    def postSync(self):
        for i,c in enumerate(self.children):
            self.ui.tab(i, text=c.label)
        return super().postSync()

class Tab(PUINode):
    pui_virtual = True
    def __init__(self, label):
        super().__init__()
        self.label = label




from .. import *
from .base import *

class TextField(QtBaseWidget):
    def __init__(self, model, edit_model=None):
        super().__init__()
        self.model = model
        self.edit_model = edit_model
        self.editing = False

    def update(self, prev):
        model_value = str(self.model.value)
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.curr_value = prev.curr_value
            self.ui.node = self
            self.ui.textChanged.disconnect()
            if self.curr_value.set(model_value) and not self.editing:
                self.ui.setText(model_value)
            self.ui.textChanged.connect(self.on_textchanged)
            self.ui.editingFinished.disconnect()
            self.ui.editingFinished.connect(self.on_editing_finished)
        else:
            self.ui = QtWidgets.QLineEdit()
            self.ui.setFocusPolicy(QtCore.Qt.ClickFocus | QtCore.Qt.NoFocus)
            self.ui.node = self
            self.ui.setText(model_value)
            self.curr_value = Prop(model_value)
            self.ui.textChanged.connect(self.on_textchanged)
            self.ui.editingFinished.connect(self.on_editing_finished)

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        super().update(prev)

    def on_editing_finished(self):
        node = self.get_node()
        node.editing = False
        value = self.ui.text()
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        model_value = str(self.model.value)
        self.ui.blockSignals(True)
        self.ui.setText(model_value)
        self.ui.blockSignals(False)
        e = PUIEvent()
        e.value = value
        self._change(e)
        node.ui.clearFocus()

    def on_textchanged(self):
        node = self.get_node()
        node.editing = True
        value = self.ui.text()
        if node.edit_model:
           node.edit_model.value = value
        e = PUIEvent()
        e.value = value
        self._input(e)



from .. import *
from PySide6 import QtCore, QtWidgets, QtGui

class QtViewSignal(QtCore.QObject):
    redraw = QtCore.Signal()

def _apply_params(ui, node):
    styles = {}
    if node.style_fontsize:
        styles["font"] = f"{node.style_fontsize}pt"
    if node.style_fontfamily:
        styles["font-family"] = node.style_fontfamily
    if node.style_color:
        styles["color"] = f"#{node.style_color:06X}"
    if node.style_bgcolor:
        styles["background-color"] = f"#{node.style_bgcolor:06X}"

    style = node.qt_params.get("Style")
    if not style is None:
        ui.setStyle(style)

    HorizontalPolicy = node.qt_params.get("HorizontalPolicy")
    if not HorizontalPolicy is None:
        ui.sizePolicy().setHorizontalPolicy(HorizontalPolicy)

    VerticalPolicy = node.qt_params.get("VerticalPolicy")
    if not VerticalPolicy is None:
        ui.sizePolicy().setVerticalPolicy(VerticalPolicy)

    SizeConstraint = node.qt_params.get("SizeConstraint")
    if not SizeConstraint is None:
        ui.setSizeConstraint(SizeConstraint)

    StyleSheet = node.qt_params.get("StyleSheet", {})
    for k,v in StyleSheet.items():
        styles[k] = v

    if hasattr(ui, "setStyleSheet"):
        ui.setStyleSheet("".join([f"{k}:{v};" for k,v in styles.items()]))

    if node.layout_padding:
        ui.setContentsMargins(*trbl2ltrb(node.layout_padding))

class QtPUIView(PUIView):
    pui_virtual = True
    def __init__(self):
        super().__init__()
        self.qt_params = {}
        self._qtsignal = QtViewSignal()
        self._qtsignal.redraw.connect(self.sync, QtCore.Qt.ConnectionType.QueuedConnection) # Use QueuedConnection to prevent nested trigger

    def destroy(self, direct):
        if direct:
            if self.ui: # PUIView doesn't have ui
                self.ui.deleteLater()
        self.ui = None
        super().destroy(direct)

    def redraw(self):
        self.dirty = True
        if self.updating:
            return
        self.updating = True
        self._qtsignal.redraw.emit()

    def update(self, prev=None):
        super().update(prev)
        _apply_params(self.ui, self)

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseWidget(PUINode):
    pui_terminal = True

    def __init__(self):
        super().__init__()
        self.qt_params = {}

    def destroy(self, direct):
        if direct:
            self.ui.deleteLater()
        self.ui = None
        super().destroy(direct)

    def update(self, prev=None):
        super().update(prev)

        sizePolicy = self.ui.sizePolicy()
        if self.layout_width is not None:
            sizePolicy.setHorizontalPolicy(QtWidgets.QSizePolicy.Preferred)
        if self.layout_height is not None:
            sizePolicy.setVerticalPolicy(QtWidgets.QSizePolicy.Preferred)
        self.ui.setSizePolicy(sizePolicy)

        if not hasattr(self.ui, "origSizeHint"):
            self.ui.origSizeHint = self.ui.sizeHint
        self.ui.sizeHint = self.qtSizeHint

        _apply_params(self.ui, self)

    def qtSizeHint(self):
        node = self.get_node()
        if not node.ui:
            return QtCore.QSize(0, 0)
        sh = node.ui.origSizeHint()
        w = sh.width()
        h = sh.height()
        if not node.layout_width is None:
            w = node.layout_width
        if not node.layout_height is None:
            h = node.layout_height
        return QtCore.QSize(w, h)

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseLayout(PUINode):
    def __init__(self):
        super().__init__()
        self.qt_params = {}

    @property
    def outer(self):
        return self.ui

    @property
    def inner(self):
        return self.layout

    def destroy(self, direct):
        if direct:
            self.ui.deleteLater()
        self.layout = None
        self.ui = None
        super().destroy(direct)

    def update(self, prev=None):
        super().update(prev)
        _apply_params(self.ui, self)

    def addChild(self, idx, child):
        from .modal import Modal
        from .layout import Spacer
        if isinstance(child, Spacer):
            self.layout.insertItem(idx, child.outer)
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            params = {}
            if not child.layout_weight is None:
                params["stretch"] = child.layout_weight
            self.layout.insertWidget(idx, child.outer, **params)

    def removeChild(self, idx, child):
        from .modal import Modal
        from .layout import Spacer
        if isinstance(child, Spacer):
            self.layout.removeItem(child.outer)
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseFrame(QtBaseWidget):
    pui_terminal = False

    def destroy(self, direct):
        if direct:
            if self.ui:
                self.ui.deleteLater()
                self.ui = None

    def addChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.ui.setWidget(child.outer)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])

class QtInPui(QtBaseWidget):
    def __init__(self, widget, *args):
        self._internal_tag = str(id(widget))
        super().__init__(*args)
        self.ui = widget

    def destroy(self, direct):
        pass

class PuiInQt(QtPUIView):
    def __init__(self, ui):
        super().__init__()
        self.ui = ui
        self.ui.update()

    def addChild(self, idx, child):
        self.ui.addChild(idx, child)

    def removeChild(self, idx, child):
        self.ui.removeChild(idx, child)


from .. import *
from .base import *

class Button(WxBaseWidget):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.SetLabel(self.text)
        else:
            self.ui = wx.Button(getWindow(self.parent), label=self.text)
            self.ui.Bind(wx.EVT_BUTTON, self._clicked)
        super().update(prev)



from .. import *
from .base import *

from PySide6 import QtWidgets, QtGui
from PySide6.QtGui import QPainter, QColor
from PySide6.QtCore import QPoint

class PUIQtCanvas(QtWidgets.QWidget):
    def __init__(self, node, width=None, height=None):
        self.node = node
        self.width = width
        self.height = height
        super().__init__()

    def minimumSizeHint(self):
        return QtCore.QSize(self.width, self.height)

    def mouseDoubleClickEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        self.node._dblclicked(e)

    def mousePressEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        self.node._mousedown(e)

    def mouseReleaseEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        self.node._mouseup(e)

    def mouseMoveEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        self.node._mousemove(e)

    def wheelEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        e.y_delta = event.pixelDelta().y()
        e.x_delta = event.pixelDelta().x()
        e.v_delta = event.angleDelta().y()
        e.h_delta = event.angleDelta().x()
        self.node._wheel(e)

    def paintEvent(self, event):
        node = self.node.get_node()
        node.qpainter = QPainter()
        node.qpainter.begin(self)
        node.qpainter.setRenderHints(QtGui.QPainter.Antialiasing, True)

        if not node.style_bgcolor is None:
            bgBrush = QtGui.QBrush()
            bgBrush.setColor(QtGui.QColor(node.style_bgcolor))
            bgBrush.setStyle(QtCore.Qt.SolidPattern)
            rect = QtCore.QRect(0, 0, self.width or self.geometry().width(), self.height or self.geometry().height())
            node.qpainter.fillRect(rect, bgBrush)

        node.painter(node, *node.args)

        node.qpainter.end()
        node.qpainter = None

class Canvas(QtBaseWidget):
    def __init__(self, painter, *args):
        super().__init__()
        self.ui = None
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.puinode = self
            self.ui.width = self.layout_width or 0
            self.ui.height = self.layout_height or 0
        else:
            self.ui = PUIQtCanvas(self, self.layout_width or 0, self.layout_height or 0)
        self.ui.setMouseTracking(bool(self._onMouseMove))
        self.ui.update()
        super().update(prev)

    def drawText(self, x, y, text, w=None, h=None, size=12, color=None, rotate=0, anchor=Anchor.LEFT_TOP):
        if w is None:
            w = self.ui.geometry().width()
        if h is None:
            h = self.ui.geometry().height()
        self.qpainter.save()
        br = self.qpainter.boundingRect(0, 0, w, h, 0, text)
        self.qpainter.restore()

        dx = 0
        dy = 0
        if anchor.value[0]=="center":
            dx = br.width()/2
        elif anchor.value[0]=="right":
            dx = br.width()

        if anchor.value[1]=="center":
            dy = br.height()/2
        elif anchor.value[1]=="bottom":
            dy = br.height()

        self.qpainter.save()
        self.qpainter.translate(int(x+dx), int(y+dy))
        self.qpainter.rotate(rotate)
        self.qpainter.translate(int(-dx), int(-dy))

        if size is not None:
            font = self.qpainter.font()
            font.setPointSize(size)
            self.qpainter.setFont(font)

        if color is not None:
            self.qpainter.setPen(QColor(color))

        self.qpainter.drawText(0, 0, w, h, 0, text)
        self.qpainter.restore()

    def drawLine(self, x1, y1, x2, y2, color=None, width=1):
        self.qpainter.save()

        pen = self.qpainter.pen()
        if color is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(color))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawLine(x1, y1, x2, y2)

        self.qpainter.restore()

    def drawPolyline(self, coords, color=None, width=1):
        self.qpainter.save()

        pen = self.qpainter.pen()
        if color is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(color))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawPolyline([QtCore.QPointF(x,y) for x,y in coords])

        self.qpainter.restore()

    def drawPolygon(self, coords, fill=None, stroke=None, width=1):
        self.qpainter.save()

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        polygon = QtGui.QPolygonF()
        for p in coords:
            polygon.append(QtCore.QPointF(*p))
        self.qpainter.drawPolygon(polygon)

        self.qpainter.restore()

    def drawRect(self, x1, y1, x2, y2, fill=None, stroke=None, width=1):
        self.qpainter.save()

        x = min(x1, x2)
        y = min(y1, y2)
        w = abs(x2-x1)
        h = abs(y2-y1)

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawRect(x, y, w, h)

        self.qpainter.restore()

    def drawEllipse(self, x, y, rx, ry, fill=None, stroke=None, width=1):
        self.qpainter.save()

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawEllipse(QtCore.QPointF(x,y), rx, ry)

        self.qpainter.restore()



from .. import *
from .base import *

class Label(TBase):
    def __init__(self, text, selectable=False):
        super().__init__()
        self.widget = None
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.widget = prev.widget
        else:
            self.ui = containers.Container()
            self.ui.set_styles("width: auto; height: auto;")
        if self._onClicked:
            if self.widget is None or not isinstance(self.widget, widgets.Button):
                if self.widget:
                    self.widget.remove()
                self.widget = widgets.Button(self.text)
                self.widget.set_styles("height: auto; border-top: none; border-bottom: none;")
                self.widget.puinode = self
            else:
                self.widget.label = self.text
        else:
            if self.widget is None or not isinstance(self.widget, widgets.Label):
                if self.widget:
                    self.widget.remove()
                self.widget = widgets.Label(self.text, markup=False)
            else:
                self.widget.update(self.text)
        super().update(prev)

    def postUpdate(self):
        self.ui.mount(self.widget)
